﻿'
' Haftungsausschluss
' Die Software wurde mit grösster Sorgfalt entwickelt. Es kann aber nicht garantiert werden,
' dass die Software auf jedem Zielsystem einwandfrei läuft. Eine vollständig fehlerfreie Software
' ist nach dem heutigen Stand der Technik nicht möglich.
' Die Software wird wie sie ist ("as is") zur Verfügung gestellt, ohne jede Garantie für Brauchbarkeit
' für einen bestimmten Anwendungsfall. Für Schäden, die durch die Nutzung (direkt oder indirekt) entstehen
' besteht keinerlei Haftung.
'
' Urheberrecht
' Alle erwähnten Marken- und Produktnamen sind Warenzeichen der jeweiligen Rechtsinhaber und gelten uneingeschränkt.
' Das Fehlen einer entsprechenden Kennzeichnung in der Software und der Dokumentation bedeutet nicht,
' dass es sich um einen freien Namen im Sinne der Waren- und Markenzeichen - Gesetzgebung handelt.
'
' Copyright
' Die Software wurde für den Verband Schweizerischer Fachhäuser für Medizinaltechnik (VSFM) entwickelt.
' Weitere Informationen sind unter www.vsfm.info oder www.smeex.ch zu finden.
' © 2008-2011 VSFM - Alle Rechte vorbehalten.
'
'

Imports Vsfm.Smeex.Framework.Core
Imports System.Runtime.InteropServices
Imports Vsfm.Smeex.Framework.Common
Imports System.Globalization
Imports System.ComponentModel
Imports Vsfm.Smeex.Framework.Core.Interfaces
Imports Vsfm.Smeex.Framework.Core.UI.ErrorUI
Imports System.Windows.Forms
Imports Microsoft.VisualBasic.Devices

#Region "Delegates"
Public Delegate Sub ProfileNameEventHandler(ByVal profileName As String)
Public Delegate Sub PercentEventHandler(ByVal percent As Integer)
Public Delegate Sub FileNameEventHandler(ByVal profileName As String, ByVal file As String)
Public Delegate Sub PlugInNameEventHandler(ByVal plugInName As String, ByVal profileName As String, ByVal file As String)
#End Region

''' <summary>
''' COM sichtbare Einstiegsklasse für die Demo Integration.
''' </summary>
<ComVisible(True), ComClass(FrameworkWrapper.ClassId, FrameworkWrapper.InterfaceId, FrameworkWrapper.EventsId)> _
Public Class FrameworkWrapper
  Implements IDisposable

#Region "Variablen"
#Region "COM-GUIDs"
  ' Diese GUIDs stellen die COM-Identität für diese Klasse 
  ' und ihre COM-Schnittstellen bereit. Wenn Sie sie ändern, können vorhandene 
  ' Clients nicht mehr auf die Klasse zugreifen.
  Public Const ClassId As String = "a79e6746-26ad-4456-9e14-74a6ef3b2a3b"
  Public Const InterfaceId As String = "913dae31-c7b5-4c98-86cd-7cff4d638ad8"
  Public Const EventsId As String = "e7736c61-057e-4b9a-9c14-bd62db7d4a45"
#End Region

  ''' <summary>
  ''' Liste von EventSink für LateBinding-Aufrufe.
  ''' </summary>
  ''' <remarks></remarks>
  Private m_EventSinks As New ArrayList
  ''' <summary>
  ''' Smeex-Framework-Instanz.
  ''' </summary>
  Private WithEvents m_Framework As SmeexFramework
  ''' <summary>
  ''' Pfad zur Access-Datenbank.
  ''' </summary>
  Private m_Database As String
#End Region

#Region "Konstruktoren"
  ' Eine erstellbare COM-Klasse muss eine Public Sub New() 
  ' ohne Parameter aufweisen. Andernfalls wird die Klasse 
  ' nicht in der COM-Registrierung registriert und kann nicht 
  ' über CreateObject erstellt werden.
  Public Sub New()
    MyBase.New()
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "New FrameworkWrapper")
  End Sub

  <System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Performance", "CA1810:InitializeReferenceTypeStaticFieldsInline")> _
  Shared Sub New()
    ErrorDialog.RegisterExportHandler(New FileExport())
    AddHandler AppDomain.CurrentDomain.UnhandledException, AddressOf OnUnhandledException
    AddHandler Application.ThreadException, AddressOf OnThreadException
    AddHandler My.Computer.Network.NetworkAvailabilityChanged, AddressOf OnNetworkAvailabilityChanged
  End Sub
#End Region

#Region "Properties"
  ''' <summary>
  ''' Gibt den Pfad zur Access-Datenbank zurück, oder legt diesen fest.
  ''' </summary>
  ''' <remarks>
  ''' Der Pfad kann nach dem erstmaligen Verbindungsaufbau nicht mehr geändert werden!
  ''' </remarks>
  Public Property Database() As String
    Get
      Database = m_Database
    End Get
    Set(ByVal value As String)
      If (m_Framework IsNot Nothing) Then Throw New InvalidOperationException()

      If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, String.Format(CultureInfo.InvariantCulture, My.Resources.FrameworkWrapper_Log_SetDatabase, value))
      m_Database = value
    End Set
  End Property

  ''' <summary>
  ''' SMEEX-Framework-Instanz.
  ''' </summary>
  Private ReadOnly Property Framework() As SmeexFramework
    Get
      If (m_Framework Is Nothing) Then
        m_Framework = New SmeexFramework(New StorageHandler(Database))
      End If
      Framework = m_Framework
    End Get
  End Property

  ''' <summary>
  ''' Legt den Log-Level für das Framework fest.
  ''' </summary>
  Public Property LogLevel() As LogLevels
    Get
      LogLevel = CType(LogFacade.Level, LogLevels)
    End Get
    Set(ByVal value As LogLevels)
      LogFacade.Level = CType(value, SourceLevels)
    End Set
  End Property
#End Region

#Region "Methoden"
  ''' <summary>
  ''' Startet die Profil-Verwaltung.
  ''' </summary>
  ''' <param name="handle">Fenster-Handle</param>
  Public Sub ManageProfileConfigurations(ByVal handle As Integer)
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "ManageProfileConfigurations")
    Try
      Framework.ManageProfileConfigurations(New Win32Handle(handle))
    Catch ex As Exception
      LogFacade.LogError(ex)
      Throw
    End Try
  End Sub

  ''' <summary>
  ''' Startet einen Export-Vorgang.
  ''' </summary>
  ''' <param name="handle">Fenster-Handle</param>
  ''' <param name="profile">zu verwendendes Profil</param>
  Public Sub Export(ByVal handle As Integer, ByVal profile As String)
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "Export")
    ExportEx(handle, profile, Nothing)
  End Sub

  ''' <summary>
  ''' Startet einen Export-Vorgang.
  ''' </summary>
  ''' <param name="handle">Fenster-Handle</param>
  ''' <param name="profile">zu verwendendes Profil</param>
  ''' <param name="params">Parameter-Array</param>
  Public Sub ExportEx(ByVal handle As Integer, ByVal profile As String, ByVal params() As String)
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "ExportEx")
    Dim prof As ProfileConfiguration

    Try
      ' Profil laden
      prof = InitializeProfile(profile, params)

      ' Export durchführen
      SmeexIdMapping.ClearCache()
      Framework.Export(New Win32Handle(handle), prof)
    Catch ex As Exception
      LogFacade.LogError(ex)
      Throw
    End Try
  End Sub

  ''' <summary>
  ''' Startet einen Import-Vorgang.
  ''' </summary>
  ''' <param name="handle">Fenster-Handle</param>
  ''' <param name="profile">zu verwendendes Profil</param>
  Public Sub Import(ByVal handle As Integer, ByVal profile As String)
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "Import")
    ImportEx(handle, profile, Nothing)
  End Sub

  ''' <summary>
  ''' Startet einen Import-Vorgang.
  ''' </summary>
  ''' <param name="handle">Fenster-Handle</param>
  ''' <param name="profile">zu verwendendes Profil</param>
  ''' <param name="params">Parameter-Array</param>
  Public Sub ImportEx(ByVal handle As Integer, ByVal profile As String, ByVal params() As String)
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "ImportEx")
    Dim prof As ProfileConfiguration

    Try
      ' Profil laden
      prof = InitializeProfile(profile, params)

      ' Export durchführen
      SmeexIdMapping.ClearCache()
      Framework.Import(New Win32Handle(handle), prof)
    Catch ex As Exception
      LogFacade.LogError(ex)
      Throw
    End Try
  End Sub

  ''' <summary>
  ''' Registriert eine EventSink, falls der Wrapper über LateBinding verwendet wird.
  ''' </summary>
  ''' <param name="eventSink">Objekt, dass als EventSink dienen soll</param>
  ''' <remarks>
  ''' Das übergebene Objekte muss eine Methode mit folgender Signatur aufweisen:
  ''' <code>
  ''' Public Sub EventRaised(eventName As String, eventParameter As String)
  ''' </code>
  ''' </remarks>
  Public Sub RegisterEventSink(ByVal eventSink As Object)
    If (eventSink IsNot Nothing) AndAlso (Not m_EventSinks.Contains(eventSink)) Then
      m_EventSinks.Add(eventSink)
    End If
  End Sub

  ''' <summary>
  ''' Entfernt die Registrierung für eine EventSink.
  ''' </summary>
  ''' <param name="eventSink">zu entfernende EventSink</param>
  ''' <remarks><seealso cref="RegisterEventSink"/></remarks>
  Public Sub UnregisterEventSink(ByVal eventSink As Object)
    If (eventSink IsNot Nothing) AndAlso (m_EventSinks.Contains(eventSink)) Then
      m_EventSinks.Remove(eventSink)
    End If
  End Sub
#End Region

#Region "Hilfsmethoden"
  ''' <summary>
  ''' Lädt das angegebene Profile und initialisiert die Parameter.
  ''' </summary>
  ''' <param name="profile">zu ladendes Profil</param>
  ''' <param name="params">zu initialisierende Parameter</param>
  ''' <returns>Initialisiertes Profil</returns>
  Private Function InitializeProfile(ByVal profile As String, ByVal params() As String) As ProfileConfiguration
    Dim sa() As String

    ' Profil laden
    InitializeProfile = Framework.LoadProfileConfiguration(profile)

    ' Parameter setzen
    If (params IsNot Nothing) Then
      For Each param As String In params
        If (Not String.IsNullOrEmpty(param)) Then
          sa = param.Split(New String() {"="}, 2, StringSplitOptions.None)
          If (sa.Length <> 2) Then Throw New ArgumentOutOfRangeException(param)
          If (Not String.IsNullOrEmpty(sa(0))) Then InitializeProfile.Parameter(sa(0)) = sa(1)
        End If
      Next
    End If
  End Function
#End Region

#Region " IDisposable Support "
  Private disposedValue As Boolean     ' So ermitteln Sie überflüssige Aufrufe

  ' IDisposable
  Protected Overridable Sub Dispose(ByVal disposing As Boolean)
    If Not Me.disposedValue Then
      If disposing Then
        ' Anderen Zustand freigeben (verwaltete Objekte).
        If (m_Framework IsNot Nothing) Then
          DirectCast(m_Framework.Storage, IDisposable).Dispose()
          m_Framework = Nothing
        End If
      End If

      ' Eigenen Zustand freigeben (nicht verwaltete Objekte).
      ' Große Felder auf NULL festlegen.
    End If
    Me.disposedValue = True
  End Sub

  ' Dieser Code wird von Visual Basic hinzugefügt, um das Dispose-Muster richtig zu implementieren.
  ''' <summary>
  ''' Rufen Sie diese Methode auf, wenn Sie alle von diesem Objekt benutzten Resourcen freigeben wollen.
  ''' </summary>
  Public Sub Dispose() Implements IDisposable.Dispose
    ' Ändern Sie diesen Code nicht. Fügen Sie oben in Dispose(ByVal disposing As Boolean) Bereinigungscode ein.
    Dispose(True)
    GC.SuppressFinalize(Me)
  End Sub
#End Region

#Region "Framework-Events"
  ''' <summary>
  ''' Event wird ausgelöst, wenn ein Export beendet wurde.
  ''' </summary>
  Public Event ExportFinished As ProfileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn ein Export gestartet wurde.
  ''' </summary>
  Public Event ExportStarted As ProfileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn ein Import beendet wurde.
  ''' </summary>
  Public Event ImportFinished As ProfileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn ein Import gestartet wurde.
  ''' </summary>
  Public Event ImportStarted As ProfileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn die Fortschrittsanzeige aktualisiert werden soll.
  ''' </summary>
  Public Event ProgressChanged As PercentEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn mit der Kommunikation begonnen wurde.
  ''' </summary>
  Public Event CommunicationStarted As FileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn die Kommunikation abgeschlossen wurde.
  ''' </summary>
  Public Event CommunicationFinished As FileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn mit der Transformation begonnen wurde.
  ''' </summary>
  Public Event TransformationStarted As FileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, wenn die Transformation abgeschlossen wurde.
  ''' </summary>
  Public Event TransformationFinished As FileNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, bevor ein PlugIn die Verarbeitung startet.
  ''' </summary>
  Public Event PlugInStarted As PlugInNameEventHandler
  ''' <summary>
  ''' Event wird ausgelöst, nachdem ein PlugIn die Verarbeitung abgeschlossen hat.
  ''' </summary>
  Public Event PlugInFinished As PlugInNameEventHandler

  Private Sub CallEventSinks(ByVal eventName As String, ByVal eventParam As Object)
    For Each obj As Object In m_EventSinks
      If (obj IsNot Nothing) Then
        Try
          CallByName(obj, "EventRaised", CallType.Method, eventName, CStr(eventParam))
        Catch ex As MemberAccessException
          LogFacade.LogWarn(c_TraceSource, String.Format(CultureInfo.InvariantCulture, My.Resources.FrameworkWrapper_Log_InvalidEventSink, obj))
        End Try
      End If
    Next
  End Sub

  Private Sub OnExportFinished(ByVal profileName As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnExportFinished")
    RaiseEvent ExportFinished(profileName)
    CallEventSinks("ExportFinished", profileName)
  End Sub

  Private Sub OnExportStarted(ByVal profileName As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnExportStarted")
    RaiseEvent ExportStarted(profileName)
    CallEventSinks("ExportStarted", profileName)
  End Sub

  Private Sub OnImportFinished(ByVal profileName As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnImportFinished")
    RaiseEvent ImportFinished(profileName)
    CallEventSinks("ImportFinished", profileName)
  End Sub

  Private Sub OnImportStarted(ByVal profileName As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnImportStarted")
    RaiseEvent ImportStarted(profileName)
    CallEventSinks("ImportStarted", profileName)
  End Sub

  Private Sub OnProgressChanged(ByVal percent As Integer)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnProgressChanged")
    RaiseEvent ProgressChanged(percent)
    CallEventSinks("ProgressChanged", percent)
  End Sub

  Private Sub OnCommunicationStarted(ByVal profileName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnCommunicationStarted")
    RaiseEvent CommunicationStarted(profileName, file)
    CallEventSinks("CommunicationStarted", String.Format(CultureInfo.InvariantCulture, "{0};{1}", profileName, file))
  End Sub

  Private Sub OnCommunicationFinished(ByVal profileName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnCommunicationFinished")
    RaiseEvent CommunicationFinished(profileName, file)
    CallEventSinks("CommunicationFinished", String.Format(CultureInfo.InvariantCulture, "{0};{1}", profileName, file))
  End Sub

  Private Sub OnTransformationStarted(ByVal profileName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnTransformationStarted")
    RaiseEvent TransformationStarted(profileName, file)
    CallEventSinks("TransformationStarted", String.Format(CultureInfo.InvariantCulture, "{0};{1}", profileName, file))
  End Sub

  Private Sub OnTransformationFinished(ByVal profileName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnTransformationFinished")
    RaiseEvent TransformationFinished(profileName, file)
    CallEventSinks("TransformationFinished", String.Format(CultureInfo.InvariantCulture, "{0};{1}", profileName, file))
  End Sub

  Private Sub OnPlugInStarted(ByVal profileName As String, ByVal plugInName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnPlugInStarted")
    RaiseEvent PlugInStarted(plugInName, profileName, file)
    CallEventSinks("PlugInStarted", String.Format(CultureInfo.InvariantCulture, "{0};{1};{2}", profileName, plugInName, file))
  End Sub

  Private Sub OnPlugInFinished(ByVal profileName As String, ByVal plugInName As String, ByVal file As String)
    If (LogFacade.IsDebugEnabled) Then LogFacade.LogDebug(c_TraceSource, "OnPlugInFinished")
    RaiseEvent PlugInFinished(plugInName, profileName, file)
    CallEventSinks("PlugInFinished", String.Format(CultureInfo.InvariantCulture, "{0};{1};{2}", profileName, plugInName, file))
  End Sub

#Region "Event-Weiterleitung"
  Private Sub _ExportFinished(ByVal sender As Object, ByVal e As ProfileEventArgs) Handles m_Framework.ExportFinished
    OnExportFinished(e.Profile.Name)
  End Sub

  Private Sub _ExportStarted(ByVal sender As Object, ByVal e As ProfileEventArgs) Handles m_Framework.ExportStarted
    OnExportStarted(e.Profile.Name)
  End Sub

  Private Sub _ImportFinished(ByVal sender As Object, ByVal e As ProfileEventArgs) Handles m_Framework.ImportFinished
    OnImportFinished(e.Profile.Name)
  End Sub

  Private Sub _ImportStarted(ByVal sender As Object, ByVal e As ProfileEventArgs) Handles m_Framework.ImportStarted
    OnImportStarted(e.Profile.Name)
  End Sub

  Private Sub _ProgressChanged(ByVal sender As Object, ByVal e As ProgressChangedEventArgs) Handles m_Framework.ProgressChanged
    OnProgressChanged(e.ProgressPercentage)
  End Sub

  Private Sub _CommunicationStarted(ByVal sender As Object, ByVal e As RegionEventArgs) Handles m_Framework.CommunicationStarted
    OnCommunicationStarted(e.Profile.Name, e.File)
  End Sub

  Private Sub _CommunicationFinished(ByVal sender As Object, ByVal e As RegionEventArgs) Handles m_Framework.CommunicationFinished
    OnCommunicationFinished(e.Profile.Name, e.File)
  End Sub

  Private Sub _TransformationStarted(ByVal sender As Object, ByVal e As RegionEventArgs) Handles m_Framework.TransformationStarted
    OnTransformationStarted(e.Profile.Name, e.File)
  End Sub

  Private Sub _TransformationFinished(ByVal sender As Object, ByVal e As RegionEventArgs) Handles m_Framework.TransformationFinished
    OnTransformationFinished(e.Profile.Name, e.File)
  End Sub

  Private Sub _PlugInStarted(ByVal sender As Object, ByVal e As PlugInEventArgs) Handles m_Framework.PlugInStarted
    OnPlugInStarted(e.Profile.Name, e.PlugIn.Name, e.File)
  End Sub

  Private Sub _PlugInFinished(ByVal sender As Object, ByVal e As PlugInEventArgs) Handles m_Framework.PlugInFinished
    OnPlugInFinished(e.Profile.Name, e.PlugIn.Name, e.File)
  End Sub
#End Region
#End Region

#Region "Globale Events"
  Private Shared Sub OnUnhandledException(ByVal sender As Object, ByVal e As UnhandledExceptionEventArgs)
    Dim ex As Exception = Nothing
    Dim s As String = String.Empty

    If GetType(Exception).IsInstanceOfType(e.ExceptionObject) Then
      ex = DirectCast(e.ExceptionObject, Exception)
    Else
      s = " " + e.ExceptionObject.ToString()
    End If

    If (ex IsNot Nothing) Then s = ex.Message
    LogFacade.LogCritical(c_TraceSource, s)
    UIFactory.Singleton.ErrorDialog.ShowError(s, ex)

    'Programm beenden, da sonst die Fehlermeldung erscheint: "ApplikationNNN hat ein Problem festgestellt und muss beendet werden."
    Environment.Exit(-1)
  End Sub

  Private Shared Sub OnThreadException(ByVal sender As Object, ByVal e As Threading.ThreadExceptionEventArgs)
    LogFacade.LogCritical(c_TraceSource, e.Exception, e.Exception.Message)
    UIFactory.Singleton.ErrorDialog.ShowError(e.Exception.Message, e.Exception)
  End Sub

  Private Shared Sub OnNetworkAvailabilityChanged(ByVal sender As Object, ByVal e As NetworkAvailableEventArgs)
    LogFacade.LogWarn(String.Format(CultureInfo.InvariantCulture, My.Resources.FrameworkWrapper_Log_NetworkChange, e.IsNetworkAvailable))
  End Sub
#End Region
End Class

#Region "Enum"
''' <summary>
''' Enumeration der Framework-LogLevel.
''' </summary>
<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1008:EnumsShouldHaveZeroValue")> _
<System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2217:DoNotMarkEnumsWithFlags")> _
<Flags(), ComVisible(True)> _
Public Enum LogLevels
  ActivityTracing = &HFF00
  All = -1
  [Off] = 0
  Critical = 1
  [Error] = 3
  Information = 15
  Warning = 7
  Verbose = &H1F
End Enum
#End Region