﻿'
' Haftungsausschluss
' Die Software wurde mit grösster Sorgfalt entwickelt. Es kann aber nicht garantiert werden,
' dass die Software auf jedem Zielsystem einwandfrei läuft. Eine vollständig fehlerfreie Software
' ist nach dem heutigen Stand der Technik nicht möglich.
' Die Software wird wie sie ist ("as is") zur Verfügung gestellt, ohne jede Garantie für Brauchbarkeit
' für einen bestimmten Anwendungsfall. Für Schäden, die durch die Nutzung (direkt oder indirekt) entstehen
' besteht keinerlei Haftung.
'
' Urheberrecht
' Alle erwähnten Marken- und Produktnamen sind Warenzeichen der jeweiligen Rechtsinhaber und gelten uneingeschränkt.
' Das Fehlen einer entsprechenden Kennzeichnung in der Software und der Dokumentation bedeutet nicht,
' dass es sich um einen freien Namen im Sinne der Waren- und Markenzeichen - Gesetzgebung handelt.
'
' Copyright
' Die Software wurde für den Verband Schweizerischer Fachhäuser für Medizinaltechnik (VSFM) entwickelt.
' Weitere Informationen sind unter www.vsfm.info oder www.smeex.ch zu finden.
' © 2008-2011 VSFM - Alle Rechte vorbehalten.
'
'

Imports Vsfm.Smeex.Framework.Common
Imports System.Globalization
Imports System.Text
Imports Vsfm.Smeex.Framework.Core
Imports Microsoft.Win32

''' <summary>
''' Mapping-Klasse für SmeexId zu Datenbank-Strukturen.
''' </summary>
Friend NotInheritable Class SmeexIdMapping
#Region "Konstanten / Variablen"
  Public Const c_VsfmRoot As String = "2.16.756.5.30.1.106"
  Public Const c_SmeexRoot As String = c_VsfmRoot + ".1"

  Public Const c_ProfileConfiguration As String = c_SmeexRoot + ".1"
  Public Const c_ProfileName As String = c_ProfileConfiguration + ".1"
  Public Const c_ProfileData As String = c_ProfileConfiguration + ".2"

  Private Shared m_SmeexIdCache As IDictionary(Of String, SmeexIdStructure)
#End Region

#Region "Konstruktoren"
  Private Sub New()
  End Sub
#End Region

#Region "Methoden"
  ''' <summary>
  ''' Erstellt aus den übergebenen Daten ein SQL-Statement.
  ''' </summary>
  ''' <param name="storage">Datenbank-Zugriff</param>
  ''' <param name="profile">Profil-Instanz</param>
  ''' <param name="tbl">SMEEX-Tabellen-Definition</param>
  ''' <returns>SQL-Statement zum Laden der definierten Daten</returns>
  ''' <exception cref="NotSupportedException">Wird ausgelöst, falls eine SmeexId nicht unterstützt wird.</exception>
  Public Shared Function BuildSQL(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal tbl As Table) As String
    Dim name As String
    Dim sb As StringBuilder
    Dim sbWhere As New StringBuilder
    Dim tblName As String
    CheckCommonParameters(storage, profile, tbl)

    ' Spalten
    sb = New StringBuilder("SELECT ")
    For Each col In tbl.Columns
      If (String.Equals(col.SmeexId, tbl.SmeexId, StringComparison.Ordinal)) Then
        name = "ID"
      Else
        name = MapSmeexId(storage, col.SmeexId)
      End If
      sb.Append(name).Append(" AS ").Append(col.Name).Append(", ")
      If (profile.Parameter(col.SmeexId) IsNot Nothing) Then
        If (sbWhere.Length > 0) Then sbWhere.Append(" AND ")
        sbWhere.Append("(").Append(name).Append(" = {").Append(col.SmeexId).Append("})")
      End If
    Next
    sb.Remove(sb.Length - 2, 2)

    ' Tabellen
    tblName = MapSmeexId(storage, tbl.SmeexId)
    sb.Append(" FROM ").Append(tblName).Append(" AS ").Append(tbl.Name)

    ' JOIN
    name = ElementJoin(storage, tbl.SmeexId)
    If (Not String.IsNullOrEmpty(name)) Then
      name = name.Replace(tblName + ".", tbl.Name + ".")
      sb.Append(" ").Append(name)
    End If

    ' WHERE
    If (sbWhere.Length > 0) Then
      sbWhere.Insert(0, " WHERE ")
      sb.Append(sbWhere)
    End If

    If (Not String.IsNullOrEmpty(SQLBefore(storage, tbl.SmeexId))) Then
      sb.Insert(0, SQLBefore(storage, tbl.SmeexId) + vbCrLf + "GO" + vbCrLf)
    End If
    If (Not String.IsNullOrEmpty(SQLAfter(storage, tbl.SmeexId))) Then
      sb.Append(vbCrLf + "GO" + vbCrLf + SQLAfter(storage, tbl.SmeexId))
    End If

    ' Parameter aufbereiten
    BuildSQL = ReplaceParameterNames(sb.ToString, profile, tbl)
  End Function

  ''' <summary>
  ''' Löscht den SmeexId-Cache und lädt diesen bei der nächsten Verwendung neu.
  ''' </summary>
  Public Shared Sub ClearCache()
    m_SmeexIdCache = Nothing
  End Sub

  ''' <summary>
  ''' Fügt die Anhänge dem Archiv hinzu.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="profile">aktuelles Profil</param>
  ''' <param name="arch">Archiv-Instanz</param>
  ''' <param name="tableDefinition">Tabellen-Definition</param>
  ''' <param name="currentData">aktuelle Datensätze</param>
  Public Shared Sub AddAnnex(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal tableDefinition As Table, ByVal currentData As DataTable)
    Dim keyName As String
    Dim typeName As String
    Dim dt As DataTable
    CheckCommonParameters(storage, profile, tableDefinition)
    If (arch Is Nothing) Then Throw New ArgumentNullException("arch")
    If (currentData Is Nothing) Then Throw New ArgumentNullException("currentData")

    'TODO: An Strukturen der DemoDB-Anpassen
    If (tableDefinition.SmeexId = "0.0.0.0") Then
      keyName = tableDefinition.Columns("0.0.0.0.0").Name
      typeName = tableDefinition.Columns("0.0.0.0.13").Name
      For Each dr As DataRow In currentData.Rows
        If (String.Equals("p", CStr(dr(typeName)), StringComparison.OrdinalIgnoreCase)) Then
          AddAnnex(storage, arch, MapSmeexId(storage, "0.0.0.0"), MapSmeexId(storage, "0.0.0.0.0"), CStr(dr(keyName)))
        End If
      Next
    ElseIf (tableDefinition.SmeexId = "2.16.756.5.30.1.106.1.10.10.103.108") Then
      dt = storage.ExecuteReader("SELECT DISTINCT ID FROM SMEEX_TempTabelle")
      For Each dr As DataRow In dt.Rows
        AddAnnex(storage, arch, "Patient", "PatientNr", CStr(dr(0)))
      Next
    End If
  End Sub

  ''' <summary>
  ''' Erstellt das Check SQL-Script.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="profile">aktuelles Profil</param>
  ''' <param name="tbl">Tabellen-Deklaration</param>
  ''' <returns>SQL-Script</returns>
  Public Shared Function BuildCheckSQL(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal tbl As Table) As String
    CheckCommonParameters(storage, profile, tbl)
    BuildCheckSQL = ReplaceParameterValues(m_SmeexIdCache(tbl.SmeexId).SQLCheck, tbl)
  End Function

  ''' <summary>
  ''' Erstellt das Insert SQL-Script.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="profile">aktuelles Profil</param>
  ''' <param name="tbl">Tabellen-Deklaration</param>
  ''' <returns>SQL-Script</returns>
  Public Shared Function BuildInsertSQL(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal tbl As Table) As String
    CheckCommonParameters(storage, profile, tbl)
    BuildInsertSQL = ReplaceParameterValues(m_SmeexIdCache(tbl.SmeexId).SQLInsert, tbl)
  End Function

  ''' <summary>
  ''' Erstellt das Update SQL-Script.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="profile">aktuelles Profil</param>
  ''' <param name="tbl">Tabellen-Deklaration</param>
  ''' <returns>SQL-Script</returns>
  Public Shared Function BuildUpdateSQL(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal tbl As Table) As String
    CheckCommonParameters(storage, profile, tbl)
    BuildUpdateSQL = ReplaceParameterValues(m_SmeexIdCache(tbl.SmeexId).SQLUpdate, tbl)
  End Function
#End Region

#Region "Hilfsmethoden"
  ''' <summary>
  ''' Mapped SmeexIds zu Datenbank-Namen.
  ''' </summary>
  ''' <param name="storage">Datenbank-Zugriff</param>
  ''' <param name="smeexId">zu mappende SmeexId</param>
  ''' <returns>Name der SmeexId in der Datenbank</returns>
  ''' <exception cref="NotSupportedException">Wird ausgelöst, falls eine SmeexId nicht unterstützt wird.</exception>
  Private Shared Function MapSmeexId(ByVal storage As StorageHandler, ByVal smeexId As String) As String
    If (m_SmeexIdCache Is Nothing) Then m_SmeexIdCache = BuildSmeexIdCache(storage)

    If (Not String.IsNullOrEmpty(smeexId) AndAlso m_SmeexIdCache.ContainsKey(smeexId)) Then
      MapSmeexId = m_SmeexIdCache(smeexId).ElementName
    Else
      Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, smeexId))
    End If
  End Function

  ''' <summary>
  ''' Gibt den Join-Bereich des Elements zurück.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="smeexId">SmeexId des zu ermittelnden Scripts</param>
  ''' <returns>Join-Bereich</returns>
  Public Shared Function ElementJoin(ByVal storage As StorageHandler, ByVal smeexId As String) As String
    If (m_SmeexIdCache Is Nothing) Then m_SmeexIdCache = BuildSmeexIdCache(storage)

    If (Not String.IsNullOrEmpty(smeexId) AndAlso m_SmeexIdCache.ContainsKey(smeexId)) Then
      ElementJoin = m_SmeexIdCache(smeexId).ElementJoin
    Else
      Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, smeexId))
    End If
  End Function

  ''' <summary>
  ''' Gibt das SQLBefore-Script zurück.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="smeexId">SmeexId des zu ermittelnden Scripts</param>
  ''' <returns>SQL-Script</returns>
  Public Shared Function SQLBefore(ByVal storage As StorageHandler, ByVal smeexId As String) As String
    If (m_SmeexIdCache Is Nothing) Then m_SmeexIdCache = BuildSmeexIdCache(storage)

    If (Not String.IsNullOrEmpty(smeexId) AndAlso m_SmeexIdCache.ContainsKey(smeexId)) Then
      SQLBefore = m_SmeexIdCache(smeexId).SQLBefore
    Else
      Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, smeexId))
    End If
  End Function

  ''' <summary>
  ''' Gibt das SQLAfter-Script zurück.
  ''' </summary>
  ''' <param name="storage">Storage-Instanz</param>
  ''' <param name="smeexId">SmeexId des zu ermittelnden Scripts</param>
  ''' <returns>SQL-Script</returns>
  Public Shared Function SQLAfter(ByVal storage As StorageHandler, ByVal smeexId As String) As String
    If (m_SmeexIdCache Is Nothing) Then m_SmeexIdCache = BuildSmeexIdCache(storage)

    If (Not String.IsNullOrEmpty(smeexId) AndAlso m_SmeexIdCache.ContainsKey(smeexId)) Then
      SQLAfter = m_SmeexIdCache(smeexId).SQLAfter
    Else
      Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, smeexId))
    End If
  End Function

  ''' <summary>
  ''' Ersetzt die Parameterwerte durch Spaltennamen im übergebenen SQL-Statement.
  ''' </summary>
  ''' <param name="sql">SQL-Statement, dessen Parameter ersetzt werden sollen</param>
  ''' <param name="profile">Profil-Instanz, die die Parameter enthält</param>
  ''' <param name="tbl">Tabellen-Definition</param>
  ''' <returns>SQL-Statement mit ersetzten Parametern</returns>
  ''' <remarks>
  ''' Neben den Profil-Parametern (<see cref="ProfileConfiguration.Parameter"/>)
  ''' werden auch SmeexId-Strings ("[smeexId]") durch die technischen Namen ersetzt
  ''' (<see cref="Table.Name"/> und <see cref="Column.Name"/>).
  ''' </remarks>
  Private Shared Function ReplaceParameterNames(ByVal sql As String, ByVal profile As ProfileConfiguration, ByVal tbl As Table) As String
    For Each key As String In profile.ParameterKeys
      If (profile.Parameter(key) IsNot Nothing) Then
        sql = sql.Replace("{" + key + "}", Convert.ToString(profile.Parameter(key), CultureInfo.InvariantCulture))
      End If
    Next

    ' SmeexIds durch effektive Namen (Tabelle oder Spalten) ersetzen
    For Each col In tbl.Columns
      sql = sql.Replace("[" + col.SmeexId + "]", col.Name)
    Next
    sql = sql.Replace("[" + tbl.SmeexId + "]", tbl.Name)

    ReplaceParameterNames = sql
  End Function

  ''' <summary>
  ''' Ersetzt die Parameterwerte durch Parameternamen im übergebenen SQL-Statement.
  ''' </summary>
  ''' <param name="sql">SQL-Statement, dessen Parameter ersetzt werden sollen</param>
  ''' <param name="tbl">Tabellen-Definition</param>
  ''' <returns>SQL-Statement mit ersetzten Parametern</returns>
  Private Shared Function ReplaceParameterValues(ByVal sql As String, ByVal tbl As Table) As String
    Dim lv As Integer
    sql = IsNull(sql, String.Empty)
    For Each col In tbl.Columns
      sql = sql.Replace("[" + col.SmeexId + "]", "@param" & lv)
      lv += 1
    Next
    ReplaceParameterValues = sql
  End Function

  ''' <summary>
  ''' Erstellt das Mapping zwischen SmeexId und Datenbank-Namen.
  ''' </summary>
  Private Shared Function BuildSmeexIdCache(ByVal storage As StorageHandler) As IDictionary(Of String, SmeexIdStructure)
    Dim dt As DataTable
    BuildSmeexIdCache = New Dictionary(Of String, SmeexIdStructure)
    dt = storage.ExecuteReader("SELECT SmeexId, ElementName, ElementJoin, SQLVorher, SQLNachher, SQLCheck, SQLInsert, SQLUpdate FROM SMEEX_IDMapping")
    For Each dr As DataRow In dt.Rows
      BuildSmeexIdCache.Add(CStr(dr("SmeexId")), New SmeexIdStructure(CStr(dr("SmeexId")), CStr(dr("ElementName")), IsNull(dr("ElementJoin"), ""), _
                                                                      IsNull(dr("SQLVorher"), ""), IsNull(dr("SQLNachher"), ""), _
                                                                      IsNull(dr("SQLCheck"), ""), IsNull(dr("SQLInsert"), ""), IsNull(dr("SQLUpdate"), "")))
    Next
  End Function

  ''' <summary>
  ''' Liefert einen Ersatzwert zurück, falls ein Null-Wert übergeben wurde.
  ''' </summary>
  ''' <param name="obj">zu prüfender Wert</param>
  ''' <param name="defaultValue">Ersatzwert, falls <paramref name="obj"/> Null ist</param>
  ''' <returns>Original- oder Ersatzwert</returns>
  ''' <remarks>Als Null-Wert gelten sowohl Nothing als auch DBNull.</remarks>
  Private Shared Function IsNull(ByVal obj As Object, ByVal defaultValue As String) As String
    If (obj Is Nothing) OrElse DBNull.Value.Equals(obj) Then obj = defaultValue
    IsNull = CStr(obj)
  End Function

  ''' <summary>
  ''' Fügt alle Anhänge hinzu.
  ''' </summary>
  Private Shared Sub AddAnnex(ByVal storage As StorageHandler, ByVal arch As Archive, ByVal table As String, ByVal keyName As String, ByVal keyValue As String)
    Dim data() As Byte
    Dim newData() As Byte
    Dim sql As String

    sql = String.Format(CultureInfo.InvariantCulture, _
                        "SELECT Bild.FileName, Bild.FileData FROM {0} WHERE ({1} = @param0) AND (NOT Bild.FileName IS NULL)", _
                        table, keyName)

    Using dt As DataTable = storage.ExecuteReader(sql, keyValue)
      For Each dr As DataRow In dt.Rows
        ' Access-Header aus Daten entfernen ([0] enthält Header-Länge)
        data = CType(dr("Bild.FileData"), Byte())
        ReDim newData(data.Length - data(0) - 1)
        Array.Copy(data, CInt(data(0)), newData, 0, newData.Length)

        ' Eintrag hinzufügen
        AddSingleAnnex(storage, arch, table, keyName, keyValue, newData, CStr(dr("Bild.FileName")))
      Next
    End Using
  End Sub

  ''' <summary>
  ''' Fügt einen einzelnen Anhang dem Archiv hinzu.
  ''' </summary>
  Private Shared Sub AddSingleAnnex(ByVal storage As StorageHandler, ByVal arch As Archive, ByVal table As String, ByVal keyName As String, ByVal keyValue As String, ByVal image As Byte(), ByVal name As String)
    Dim extension As String
    Dim fi As IO.FileInfo
    Dim dtID As DataTable
    Dim path As String

    path = IO.Path.Combine(IO.Path.Combine(IO.Path.GetTempPath, "smeex"), "DemoDB")
    If (Not IO.Directory.Exists(path)) Then IO.Directory.CreateDirectory(path)
    extension = New IO.FileInfo(name).Extension

    ' Dokumenten-Tabelle füllen
    storage.ExecuteSQL("INSERT INTO SMEEX_Dokument (Name, MimeType, Groesse, Herkunft, HerkunftID) VALUES (@param0, @param1, @param2, @param3, @param4)",
                       name, FileHelper.GetMimeType(extension), image.Length, table, keyValue)
    dtID = storage.ExecuteReader("SELECT CStr(MAX(IdentId)) FROM SMEEX_Dokument")

    fi = New IO.FileInfo(IO.Path.Combine(path, String.Format(CultureInfo.InvariantCulture, "{0}{1}", dtID.Rows(0)(0), extension)))
    My.Computer.FileSystem.WriteAllBytes(fi.FullName, image, False)
    FileHelper.RegisterTempPath(arch, fi.FullName)
    If (Not fi.Exists) Then fi = New IO.FileInfo(IO.Path.Combine(IO.Path.GetDirectoryName(storage.Database), fi.Name))
    If (Not arch.Entries.Contains(fi.Name, EntryType.Annex) AndAlso fi.Exists) Then
      arch.AddAnnex(fi, name)
    End If
  End Sub

  ''' <summary>
  ''' Prüft allgemeine Parameter.
  ''' </summary>
  ''' <param name="storage">Datenbank-Zugriff</param>
  ''' <param name="profile">Profil-Instanz</param>
  ''' <param name="tbl">SMEEX-Tabellen-Definition</param>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, falls einer der Parameter Nothing ist</exception>
  Private Shared Sub CheckCommonParameters(ByVal storage As StorageHandler, ByVal profile As ProfileConfiguration, ByVal tbl As Table)
    If (storage Is Nothing) Then Throw New ArgumentNullException("storage")
    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    If (tbl Is Nothing) Then Throw New ArgumentNullException("tbl")
    If (String.IsNullOrEmpty(tbl.SmeexId)) Then Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, tbl.SmeexId))
    If (m_SmeexIdCache Is Nothing) Then m_SmeexIdCache = BuildSmeexIdCache(storage)
    If (Not m_SmeexIdCache.ContainsKey(tbl.SmeexId)) Then Throw New NotSupportedException(String.Format(CultureInfo.InvariantCulture, My.Resources.SmeexIdMapping_Exception_SmeexIdNotSupported, tbl.SmeexId))
  End Sub
#End Region

  ''' <summary>
  ''' Hilfsstruktur für die Verwaltung der SmeexIds.
  ''' </summary>
  Private Class SmeexIdStructure
#Region "Variablen"
    Private m_SmeexId As String
    Private m_ElementName As String
    Private m_ElementJoin As String
    Private m_SQLBefore As String
    Private m_SQLAfter As String
    Private m_SQLCheck As String
    Private m_SQLInsert As String
    Private m_SQLUpdate As String
#End Region

#Region "Konstruktoren"
    Public Sub New(ByVal smeexId As String, ByVal elementName As String, ByVal elementJoin As String, ByVal sqlBefore As String, ByVal sqlAfter As String, ByVal sqlCheck As String, ByVal sqlInsert As String, ByVal sqlUpdate As String)
      m_SmeexId = smeexId
      m_ElementName = elementName
      m_ElementJoin = elementJoin
      m_SQLBefore = sqlBefore
      m_SQLAfter = sqlAfter
      m_SQLCheck = sqlCheck
      m_SQLInsert = sqlInsert
      m_SQLUpdate = sqlUpdate
    End Sub
#End Region

#Region "Properties"
    ''' <summary>
    ''' SmeexId des Eintrags.
    ''' </summary>
    Public ReadOnly Property SmeexId() As String
      Get
        SmeexId = m_SmeexId
      End Get
    End Property

    ''' <summary>
    ''' Element-Name in der Datenbank.
    ''' </summary>
    Public ReadOnly Property ElementName() As String
      Get
        ElementName = m_ElementName
      End Get
    End Property

    ''' <summary>
    ''' Join-Bereich des Elements.
    ''' </summary>
    Public ReadOnly Property ElementJoin() As String
      Get
        ElementJoin = m_ElementJoin
      End Get
    End Property

    ''' <summary>
    ''' SQL-Script, das vor dem Laden von Daten (Export) ausgeführt werden soll.
    ''' </summary>
    Public ReadOnly Property SQLBefore() As String
      Get
        SQLBefore = m_SQLBefore
      End Get
    End Property

    ''' <summary>
    ''' SQL-Script, das nach dem Laden von Daten (Export) ausgeführt werden soll.
    ''' </summary>
    Public ReadOnly Property SQLAfter() As String
      Get
        SQLAfter = m_SQLAfter
      End Get
    End Property

    ''' <summary>
    ''' SQL-Script zur Prüfung, ob ein Datensatz bereits vorhanden ist (Import).
    ''' </summary>
    Public ReadOnly Property SQLCheck() As String
      Get
        SQLCheck = m_SQLCheck
      End Get
    End Property

    ''' <summary>
    ''' SQL-Script für den Insert eines neuen Datensatzes (Import).
    ''' </summary>
    Public ReadOnly Property SQLInsert() As String
      Get
        SQLInsert = m_SQLInsert
      End Get
    End Property

    ''' <summary>
    ''' SQL-Script für den Update eines vorhandenen Datensatzes (Import).
    ''' </summary>
    Public ReadOnly Property SQLUpdate() As String
      Get
        SQLUpdate = m_SQLUpdate
      End Get
    End Property
#End Region
  End Class
End Class
