﻿'
' Haftungsausschluss
' Die Software wurde mit grösster Sorgfalt entwickelt. Es kann aber nicht garantiert werden,
' dass die Software auf jedem Zielsystem einwandfrei läuft. Eine vollständig fehlerfreie Software
' ist nach dem heutigen Stand der Technik nicht möglich.
' Die Software wird wie sie ist ("as is") zur Verfügung gestellt, ohne jede Garantie für Brauchbarkeit
' für einen bestimmten Anwendungsfall. Für Schäden, die durch die Nutzung (direkt oder indirekt) entstehen
' besteht keinerlei Haftung.
'
' Urheberrecht
' Alle erwähnten Marken- und Produktnamen sind Warenzeichen der jeweiligen Rechtsinhaber und gelten uneingeschränkt.
' Das Fehlen einer entsprechenden Kennzeichnung in der Software und der Dokumentation bedeutet nicht,
' dass es sich um einen freien Namen im Sinne der Waren- und Markenzeichen - Gesetzgebung handelt.
'
' Copyright
' Die Software wurde für den Verband Schweizerischer Fachhäuser für Medizinaltechnik (VSFM) entwickelt.
' Weitere Informationen sind unter www.vsfm.info oder www.smeex.ch zu finden.
' © 2008-2011 VSFM - Alle Rechte vorbehalten.
'
'

Imports Vsfm.Smeex.Framework.Core.Interfaces
Imports Vsfm.Smeex.Framework.Core
Imports Vsfm.Smeex.Framework.Common
Imports System.Data.OleDb
Imports System.Data.Common
Imports System.Globalization
Imports System.Text
Imports System.Text.RegularExpressions
Imports System.Reflection

''' <summary>
''' Storage-Handler für den Zugriff auf die Access DemoDB.
''' </summary>
Friend Class StorageHandler
  Inherits TraceListener
  Implements IStorage, IDisposable

#Region "Konstanten / Variablen"
  ''' <summary>
  ''' Root-SmeexID der DemoDB.
  ''' </summary>
  Private Const c_RootSmeexId As String = "2.16.756.5.30.1.106.1.11.1.99"
  ''' <summary>
  ''' Schlüsselname des ProtokollID-Parameters.
  ''' </summary>
  Private Const c_ProtocolParameter As String = "DemoDB.ProtocolID"
  ''' <summary>
  ''' Regulärer Ausdruck zur Ermittlung von Parametern.
  ''' </summary>
  Private Shared ReadOnly c_ParameterRegex As Regex = New Regex("@param\d+\b")
  ''' <summary>
  ''' Regulärer Ausdruck zur Umwandlung eines friendlyNames in einen techn. Namen.
  ''' </summary>
  Private Shared ReadOnly c_InvalidCharsRegex As Regex = New Regex("([^a-zA-Z0-9])")
  ''' <summary>
  ''' Parameter-Schlüssel für das Ausschalten der Archiv-Validierung.
  ''' </summary>
  Public Const c_ParameterDisableValidation As String = "Storage.DisableValidation"
  ''' <summary>
  ''' Parameter-Schlüssel für die Änderung der Tabellen- und Spalten-Namen.
  ''' </summary>
  Public Const c_ParameterChangeNames As String = "Storage.ChangeNames"
  ''' <summary>
  ''' Aktive Verbindung zur Datenbank.
  ''' </summary>
  Private m_Connection As IDbConnection
  ''' <summary>
  ''' Access-Datenbank.
  ''' </summary>
  Private m_Database As String
  ''' <summary>
  ''' Verarbeitete Datentabellen.
  ''' </summary>
  Private m_Processed As New Dictionary(Of Integer, Object)
#End Region

#Region "Konstruktoren"
  ''' <summary>
  ''' Erstellt einen neuen StorageHandler mit Verbindung zur angegebenen Access-Datenbank.
  ''' </summary>
  ''' <param name="mdbPath">Pfad zur Access-Datenbank</param>
  Public Sub New(ByVal mdbPath As String)
    If (String.IsNullOrEmpty(mdbPath)) Then Throw New ArgumentNullException("mdbPath")
    If (Not IO.File.Exists(mdbPath)) Then Throw New IO.FileNotFoundException()
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, "New FrameworkWrapper(mdbPath)")

    m_Connection = InitializeConnection(mdbPath)
    m_Database = mdbPath
    InitializeLogging()
  End Sub
#End Region

#Region "Properties"
  ''' <summary>
  ''' Gibt die verbundene Datenbank zurück.
  ''' </summary>
  Friend ReadOnly Property Database() As String
    Get
      Database = m_Database
    End Get
  End Property
#End Region

#Region "IStorage"
  ''' <summary>
  ''' <see cref="IStorage.AddAnnex"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="profile"/> oder <paramref name="arch"/> leer ist.</exception>
  Public Sub AddAnnex(ByVal profile As ProfileConfiguration, ByVal arch As Archive) Implements IStorage.AddAnnex
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.DeleteProfileConfiguration"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="name"/> leer ist.</exception>
  Public Sub DeleteProfileConfiguration(ByVal name As String) Implements IStorage.DeleteProfileConfiguration
    If (String.IsNullOrEmpty(name)) Then Throw New ArgumentNullException("name")

    ExecuteSQL("DELETE FROM SMEEX_Profil WHERE (ProfilName = @param0)", name)
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.DeleteSetting"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="section"/> oder <paramref name="key"/> leer ist.</exception>
  Public Sub DeleteSetting(ByVal section As String, ByVal key As String) Implements IStorage.DeleteSetting
    If (String.IsNullOrEmpty(section)) Then Throw New ArgumentNullException("section")
    If (String.IsNullOrEmpty(key)) Then Throw New ArgumentNullException("key")

    section = "smeex." + section  ' Kennzeichnung der SMEEX-Sektionen
    ExecuteSQL("DELETE FROM Schalter WHERE (Sektion = @param0) AND (Schalter = @param1)", section, key)
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.InitializeNewArchive"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="profile"/> oder <paramref name="arch"/> leer ist.</exception>
  Public Sub InitializeNewArchive(ByVal profile As ProfileConfiguration, ByVal arch As Archive) Implements IStorage.InitializeNewArchive
    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    If (arch Is Nothing) Then Throw New ArgumentNullException("arch")

    ' Mindestens Sprache und Dokumenten-ID setzen
    arch.Language = "DE"
    arch.SmeexId = GetNewDocumentId()
    arch.GeneratorSmeexId = c_RootSmeexId

    ' Parameter auswerten
    If (Not String.IsNullOrEmpty(CStr(profile.Parameter(c_ParameterDisableValidation)))) Then
      arch.Validation = ValidationMode.None
    End If
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.ListProfileConfigurations"/>
  ''' </summary>
  Public Function ListProfileConfigurations() As String() Implements IStorage.ListProfileConfigurations
    ListProfileConfigurations = ToStringList(ExecuteReader("SELECT ProfilName FROM SMEEX_Profil")).ToArray
  End Function

  ''' <summary>
  ''' <see cref="IStorage.ListSettings"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="section"/> leer ist.</exception>
  Public Function ListSettings(ByVal section As String) As IList(Of String) Implements IStorage.ListSettings
    If (String.IsNullOrEmpty(section)) Then Throw New ArgumentNullException("section")

    section = "smeex." + section  ' Kennzeichnung der SMEEX-Sektionen
    ListSettings = ToStringList(ExecuteReader("SELECT Schalter FROM Schalter WHERE (Sektion = @param0)", section))
  End Function

  ''' <summary>
  ''' <see cref="IStorage.StartExport"/>
  ''' </summary>
  Public Sub StartExport(ByVal profile As ProfileConfiguration, ByVal arch As Archive) Implements IStorage.StartExport
    StartProtocoll(profile, "EXPORT")
    ExecuteSQL("DELETE FROM SMEEX_Key")
    ExecuteSQL("DELETE FROM SMEEX_Dokument")
    ExecuteSQL("DELETE FROM SMEEX_TempTabelle")
    m_Processed.Clear()
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.FinishExport"/>
  ''' </summary>
  Public Sub FinishExport(ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal success As Boolean) Implements IStorage.FinishExport
    FinishProtocoll(profile, success)
    m_Processed.Clear()
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.LoadData"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="profile"/>, <paramref name="tableDefinition"/> oder <paramref name="currentData"/> leer ist.</exception>
  Public Function LoadData(ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal tableDefinition As Table, ByVal currentData As DataTable, ByVal relationData As DataTable, ByVal sourceRelation As Relation) As DataTable Implements IStorage.LoadData
    Dim s As String
    Dim sql As String = Nothing
    Dim dt As DataTable = Nothing
    Dim data As ExportData = Nothing
    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    If (arch Is Nothing) Then Throw New ArgumentNullException("arch")
    If (tableDefinition Is Nothing) Then Throw New ArgumentNullException("tableDefinition")
    If (currentData Is Nothing) Then Throw New ArgumentNullException("currentData")

    ' Parameter-Auswertung
    CheckNames(profile, tableDefinition, currentData)

    ' doppelte Aufrufe unterdrücken
    If (Not m_Processed.ContainsKey(tableDefinition.GetHashCode)) Then
      m_Processed.Add(tableDefinition.GetHashCode, Nothing)
      s = SmeexIdMapping.BuildSQL(Me, profile, tableDefinition)
      Try
        For Each sql In s.Split(New String() {vbCrLf + "GO" + vbCrLf}, StringSplitOptions.RemoveEmptyEntries)
          If (sql.StartsWith("{", StringComparison.Ordinal)) Then
            ' Programm-Funktion ausführen
            If (data Is Nothing) Then data = New ExportData(profile, arch, tableDefinition, currentData, relationData, sourceRelation)
            ExecuteSub(data, sql)
          Else
            ' SQL-Statment ausführen
            dt = ExecuteReader(sql)
            If (dt.Rows.Count > 0) Then
              ConvertDataTypes(currentData, dt)
              currentData.Merge(dt)
              currentData.AcceptChanges()
            End If
          End If
        Next
      Catch ex As ConstraintException
        If (Not String.IsNullOrEmpty(sql)) Then ex.Data.Add("SQL", sql)
        AppendTableError(ex, currentData)
        AppendTableError(ex, dt)
        Throw
      Finally
        If (data IsNot Nothing) Then data.Dispose()
      End Try
      SmeexIdMapping.AddAnnex(Me, profile, arch, tableDefinition, currentData)
    End If
    LoadData = currentData
  End Function

  ''' <summary>
  ''' <see cref="IStorage.LoadProfileConfiguration"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="name"/> leer ist.</exception>
  Public Function LoadProfileConfiguration(ByVal name As String) As String Implements IStorage.LoadProfileConfiguration
    Dim dt As DataTable
    If (String.IsNullOrEmpty(name)) Then Throw New ArgumentNullException("name")

    dt = ExecuteReader("SELECT ProfilDaten FROM SMEEX_Profil WHERE (ProfilName = @param0)", name)
    If (dt.Rows.Count = 1) Then
      LoadProfileConfiguration = Convert.ToString(dt.Rows(0)("ProfilDaten"), CultureInfo.InvariantCulture)
    Else
      LoadProfileConfiguration = String.Empty
    End If
  End Function

  ''' <summary>
  ''' <see cref="IStorage.ReadSetting"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="section"/> oder <paramref name="key"/> leer ist.</exception>
  Public Function ReadSetting(ByVal section As String, ByVal key As String, ByVal defaultValue As String) As String Implements IStorage.ReadSetting
    Dim dt As DataTable
    If (String.IsNullOrEmpty(section)) Then Throw New ArgumentNullException("section")
    If (String.IsNullOrEmpty(key)) Then Throw New ArgumentNullException("key")

    section = "smeex." + section  ' Kennzeichnung der SMEEX-Sektionen
    dt = ExecuteReader("SELECT Wert FROM Schalter WHERE (Sektion = @param0) AND (Schalter = @param1)", section, key)
    If (dt.Rows.Count = 1) Then
      ReadSetting = Convert.ToString(dt.Rows(0)("Wert"), CultureInfo.InvariantCulture)
    Else
      ReadSetting = defaultValue
    End If
  End Function

  ''' <summary>
  ''' <see cref="IStorage.SaveProfileConfiguration"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="name"/> oder <paramref name="data"/> leer ist.</exception>
  Public Sub SaveProfileConfiguration(ByVal name As String, ByVal data As String) Implements IStorage.SaveProfileConfiguration
    Dim sql As String
    If (String.IsNullOrEmpty(name)) Then Throw New ArgumentNullException("name")
    If (String.IsNullOrEmpty(data)) Then Throw New ArgumentNullException("data")

    If (Array.IndexOf(ListProfileConfigurations, name) > -1) Then
      ' UPDATE
      sql = "UPDATE SMEEX_Profil SET ProfilDaten = @param0, UpdateID = 'smeex', UpdateDatum = Now() WHERE (ProfilName = @param1)"
      ExecuteSQL(sql, data, name)
    Else
      ' INSERT
      sql = "INSERT INTO SMEEX_Profil (ProfilName, ProfilDaten, InsertID, InsertDatum) VALUES (@param0, @param1, 'smeex', Now())"
      ExecuteSQL(sql, name, data)
    End If
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.SaveSetting"/>
  ''' </summary>
  ''' <exception cref="ArgumentNullException">Wird ausgelöst, wenn <paramref name="section"/> oder <paramref name="key"/> leer ist.</exception>
  Public Sub SaveSetting(ByVal section As String, ByVal key As String, ByVal value As String) Implements IStorage.SaveSetting
    Dim sql As String
    If (String.IsNullOrEmpty(section)) Then Throw New ArgumentNullException("section")
    If (String.IsNullOrEmpty(key)) Then Throw New ArgumentNullException("key")
    If (value Is Nothing) Then value = String.Empty

    If (ListSettings(section).IndexOf(key) > -1) Then
      ' UPDATE
      section = "smeex." + section  ' Kennzeichnung der SMEEX-Sektionen
      sql = "UPDATE Schalter SET Wert = @param0, UpdateID = 'smeex', UpdateDatum = Now() WHERE (Sektion = @param1) AND (Schalter = @param2)"
      ExecuteSQL(sql, value, section, key)
    Else
      ' INSERT
      section = "smeex." + section  ' Kennzeichnung der SMEEX-Sektionen
      sql = "INSERT INTO Schalter (Sektion, Schalter, Wert, InsertID, InsertDatum) VALUES (@param0, @param1, @param2, 'smeex', Now())"
      ExecuteSQL(sql, section, key, value)
    End If
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.StartImport"/>
  ''' </summary>
  Public Sub StartImport(ByVal profile As ProfileConfiguration, ByVal arch As Archive) Implements IStorage.StartImport
    StartProtocoll(profile, "IMPORT")
    ExecuteSQL("DELETE FROM SMEEX_Key")
    ExecuteSQL("DELETE FROM SMEEX_Dokument")
    ExecuteSQL("DELETE FROM SMEEX_TempTabelle")
    m_Processed.Clear()
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.FinishImport"/>
  ''' </summary>
  Public Sub FinishImport(ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal success As Boolean) Implements IStorage.FinishImport
    FinishProtocoll(profile, success)
    m_Processed.Clear()
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.StoreAnnex"/>
  ''' </summary>
  Public Sub StoreAnnex(ByVal profile As ProfileConfiguration, ByVal arch As Archive) Implements IStorage.StoreAnnex
  End Sub

  ''' <summary>
  ''' <see cref="IStorage.StoreData"/>
  ''' </summary>
  Public Sub StoreData(ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal tableDefinition As Table, ByVal currentData As DataTable, ByVal parentData As DataTable, ByVal parentRelation As Relation) Implements IStorage.StoreData
    Dim sqlCheck As String
    Dim sqlUpdate As String
    Dim sqlInsert As String
    Dim dt As DataTable
    Dim count As Integer
    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    If (arch Is Nothing) Then Throw New ArgumentNullException("arch")
    If (tableDefinition Is Nothing) Then Throw New ArgumentNullException("tableDefinition")
    If (currentData Is Nothing) Then Throw New ArgumentNullException("currentData")

    ' SQL-Statements laden
    sqlCheck = SmeexIdMapping.BuildCheckSQL(Me, profile, tableDefinition)
    sqlUpdate = SmeexIdMapping.BuildUpdateSQL(Me, profile, tableDefinition)
    sqlInsert = SmeexIdMapping.BuildInsertSQL(Me, profile, tableDefinition)

    ' alle Zeilen verarbeiten
    For Each dr As DataRow In currentData.Rows
      dt = ExecuteReader(sqlCheck, dr.ItemArray)
      If (dt.Rows.Count = 1) Then
        count = ExecuteSQL(sqlUpdate, dr.ItemArray)
      Else
        count = ExecuteSQL(sqlInsert, dr.ItemArray)
      End If
      If (count <= 0) Then Throw New DataException(My.Resources.StorageHandler_Exception_NoDataStored)
    Next
    'TODO: SmeexIdMapping.StoreAnnex(Me, profile, arch, tableDefinition, currentData)
  End Sub
#End Region

#Region "Hilfsmethoden"
  ''' <summary>
  ''' Stellt eine Verbindung zur übergebenen Access-Datenbank her.
  ''' </summary>
  ''' <param name="path">Access-Datenbank, zu der eine Verbindung hergestellt werden soll</param>
  ''' <returns>Geöffnete Verbindung</returns>
  Private Shared Function InitializeConnection(ByVal path As String) As IDbConnection
    Dim conBuilder As New OleDbConnectionStringBuilder()
    If (path.EndsWith(".accdb", StringComparison.OrdinalIgnoreCase)) Then
      conBuilder.Provider = "Microsoft.ACE.OLEDB.12.0"
    Else
      conBuilder.Provider = "Microsoft.Jet.OLEDB.4.0"
    End If
    conBuilder.DataSource = path
    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(c_TraceSource, String.Format(CultureInfo.InvariantCulture, "Creating new connection: {0}", conBuilder.ConnectionString))
    InitializeConnection = New OleDbConnection(conBuilder.ConnectionString)
    InitializeConnection.Open()
  End Function

  ''' <summary>
  ''' Prüft einen Profil-Parameter und ändert gegebenenfalls die Namen der Tabellen und Spalten.
  ''' </summary>
  ''' <param name="profile">aktuelle Profil-Instanz</param>
  ''' <param name="tableDefinition">Tabellen-Definition</param>
  ''' <param name="currentData">Tabelle mit aktuellen Daten</param>
  Private Shared Sub CheckNames(ByVal profile As ProfileConfiguration, ByVal tableDefinition As Table, ByVal currentData As DataTable)
    Dim dc As DataColumn
    Dim ht As Hashtable

    If (Not String.IsNullOrEmpty(CStr(profile.Parameter(c_ParameterChangeNames)))) Then
      ht = DirectCast(profile.Parameter(c_ParameterChangeNames + ".Result"), Hashtable)
      If (ht Is Nothing) Then
        ht = New Hashtable
        profile.Parameter(c_ParameterChangeNames + ".Result") = ht
      End If
      If (Not ht.ContainsKey(tableDefinition.GetHashCode)) Then
        tableDefinition.Name = ToValidName(tableDefinition.Name, tableDefinition.FriendlyName)
        currentData.TableName = tableDefinition.Name
        ht.Add(tableDefinition.GetHashCode, True)
        For Each col In tableDefinition.Columns
          dc = currentData.Columns(col.Name)
          col.Name = ToValidName(col.Name, col.FriendlyName)
          dc.ColumnName = col.Name
        Next
      End If
    End If
  End Sub

  ''' <summary>
  ''' Wandelt den angegebenen "friendlyName" in einen String um, der als techn. Name verwendet werden kann.
  ''' </summary>
  ''' <param name="oldName">zuvor verwendeter Name</param>
  ''' <param name="friendlyName">zu wandelnder Name</param>
  ''' <returns>gewandelter Name</returns>
  Private Shared Function ToValidName(ByVal oldName As String, ByVal friendlyName As String) As String
    ToValidName = c_InvalidCharsRegex.Replace(oldName + ":" + friendlyName, "_")
  End Function

  ''' <summary>
  ''' Initialisiert die Log-Umgebung.
  ''' </summary>
  Private Sub InitializeLogging()
    If (m_Connection IsNot Nothing) Then
      LogFacade.Register(Me)
    End If
  End Sub

  ''' <summary>
  ''' Erstellt eine Parameter-Instanz für die angegebenen Informationen.
  ''' </summary>
  ''' <param name="name">Name des Parameters</param>
  ''' <param name="param">Parameter-Wert</param>
  ''' <returns>Parameter-Instanz</returns>
  Private Shared Function BuildParameter(ByVal name As String, ByVal param As Object) As DbParameter
    BuildParameter = New OleDbParameter(name, param)
  End Function

  ''' <summary>
  ''' Erstellt aus der ersten Spalte der Tabelle eine String-Liste.
  ''' </summary>
  ''' <param name="dt">Tabelle, deren Daten in eine String-Liste konvertiert werden sollen</param>
  ''' <returns>String-Liste mit Daten</returns>
  ''' <remarks>Falls keine Daten vorhanden sind, wird eine leere Liste (nicht Nothing) zurückgegeben.</remarks>
  Private Shared Function ToStringList(ByVal dt As DataTable) As List(Of String)
    ToStringList = New List(Of String)
    For Each dr As DataRow In dt.Rows
      ToStringList.Add(Convert.ToString(dr(0), CultureInfo.InvariantCulture))
    Next
  End Function

  ''' <summary>
  ''' Führt eine Programm-Funktion aus.
  ''' </summary>
  ''' <param name="definition">Definition der Programm-Funktion</param>
  ''' <remarks></remarks>
  Private Sub ExecuteSub(ByVal data As ExportData, ByVal definition As String)
    Dim name As String
    Dim params() As Object

    If (LogFacade.IsInfoEnabled) Then LogFacade.LogInfo(String.Format(CultureInfo.InvariantCulture, My.Resources.StorageHandler_Log_SQLSub, definition))
    If (Not String.IsNullOrEmpty(definition)) Then
      definition = definition.Trim

      ' Format: {Sub(param1, param2)}
      name = definition.Substring(1, definition.IndexOf("("c) - 1)
      params = definition.Split("()".ToCharArray)(1).Split(","c)
      If (params IsNot Nothing) AndAlso (params.Length = 0) Then params = New Object() {data} Else params = New Object() {data, params}
      Me.GetType().GetMethod("ExecuteSub_" + name, BindingFlags.Public Or BindingFlags.NonPublic Or BindingFlags.Instance).Invoke(Me, params)
    End If
  End Sub

  ''' <summary>
  ''' Führt ein SQL-String auf der Datenbank aus, und gibt die Anzahl betroffenen Datensätze zurück.
  ''' </summary>
  ''' <param name="sql">Auszuführendes SQL-Statement</param>
  ''' <param name="parameters">Parameter des SQL-Statments</param>
  ''' <returns>Anzahl betroffene Datensätze</returns>
  ''' <remarks>
  ''' SQL-Statements sollten nicht selbst zusammengesetzt werden, da dann die Gefahr von
  ''' z.B. SQL-Injection-Angriffen besteht. Es sollen nach Möglichkeit immer Parameter verwendet werden!
  ''' </remarks>
  Friend Function ExecuteSQL(ByVal sql As String, ByVal ParamArray parameters() As Object) As Integer
    Dim start As Long = Now.Ticks

    LogSQL(sql, parameters)
    ExecuteSQL = InternalExecuteSQL(sql, parameters)
    LogSQLTiming(start, sql, parameters)
  End Function

  ''' <summary>
  ''' Führt ein SQL-String auf der Datenbank aus, und gibt die Anzahl betroffenen Datensätze zurück.
  ''' </summary>
  ''' <param name="sql">Auszuführendes SQL-Statement</param>
  ''' <param name="parameters">Parameter des SQL-Statments</param>
  ''' <returns>Anzahl betroffene Datensätze</returns>
  ''' <remarks>
  ''' SQL-Statements sollten nicht selbst zusammengesetzt werden, da dann die Gefahr von
  ''' z.B. SQL-Injection-Angriffen besteht. Es sollen nach Möglichkeit immer Parameter verwendet werden!
  ''' </remarks>
  Private Function InternalExecuteSQL(ByVal sql As String, ByVal ParamArray parameters() As Object) As Integer
    Dim lv As Integer
    Dim cmd As IDbCommand

    cmd = m_Connection.CreateCommand()
    cmd.CommandText = sql
    If (parameters IsNot Nothing) Then
      For Each obj In ReorderParameters(sql, parameters)
        If (obj IsNot Nothing) Then cmd.Parameters.Add(BuildParameter("param" & lv, obj))
        lv += 1
      Next
    End If
    Try
      InternalExecuteSQL = cmd.ExecuteNonQuery()
    Catch ex As Exception
      ex.Data("SQL") = sql
      Throw
    End Try
  End Function

  ''' <summary>
  ''' Führt eine SQL-Abfrage auf der Datenbank aus, und gibt die resultierenden Daten zurück.
  ''' </summary>
  ''' <param name="sql">Auszuführendes SQL-Statement</param>
  ''' <param name="parameters">Parameter des SQL-Statments</param>
  ''' <returns>Geladene Daten</returns>
  ''' <remarks>
  ''' SQL-Statements sollten nicht selbst zusammengesetzt werden, da dann die Gefahr von
  ''' z.B. SQL-Injection-Angriffen besteht. Es sollen nach Möglichkeit immer Parameter verwendet werden!
  ''' </remarks>
  Friend Function ExecuteReader(ByVal sql As String, ByVal ParamArray parameters() As Object) As DataTable
    Dim start As Long = Now.Ticks

    LogSQL(sql, parameters)
    ExecuteReader = InternalExecuteReader(sql, parameters)
    LogSQLTiming(start, sql, parameters)
  End Function

  ''' <summary>
  ''' Führt eine SQL-Abfrage auf der Datenbank aus, und gibt die resultierenden Daten zurück.
  ''' </summary>
  ''' <param name="sql">Auszuführendes SQL-Statement</param>
  ''' <param name="parameters">Parameter des SQL-Statments</param>
  ''' <returns>Geladene Daten</returns>
  ''' <remarks>
  ''' SQL-Statements sollten nicht selbst zusammengesetzt werden, da dann die Gefahr von
  ''' z.B. SQL-Injection-Angriffen besteht. Es sollen nach Möglichkeit immer Parameter verwendet werden!
  ''' </remarks>
  Private Function InternalExecuteReader(ByVal sql As String, ByVal ParamArray parameters() As Object) As DataTable
    Dim lv As Integer
    Dim cmd As IDbCommand

    cmd = m_Connection.CreateCommand()
    cmd.CommandText = sql
    If (parameters IsNot Nothing) Then
      For Each obj In ReorderParameters(sql, parameters)
        If (obj IsNot Nothing) Then cmd.Parameters.Add(BuildParameter("param" & lv, obj))
        lv += 1
      Next
    End If
    Try
      Using reader = cmd.ExecuteReader()
        InternalExecuteReader = New DataTable
        InternalExecuteReader.Locale = CultureInfo.InvariantCulture
        InternalExecuteReader.Load(reader)
      End Using
    Catch ex As Exception
      ex.Data("SQL") = sql
      Throw
    End Try
  End Function

  ''' <summary>
  ''' Access unterstützt keine Named-Parameter.
  ''' Deshalb muss die Reihenfolge der übergebenen Parameter
  ''' an die Reihenfolge im SQL angepasst werden.
  ''' </summary>
  ''' <param name="sql">SQL-String mit Parametern</param>
  ''' <param name="parameters">Liste der Parameter</param>
  ''' <returns>neu geordnete Liste der Parameter</returns>
  ''' <remarks>Nicht verwendete Parameter werden aus der Auflistung entfernt!</remarks>
  Private Shared Function ReorderParameters(ByVal sql As String, ByVal ParamArray parameters() As Object) As Object()
    Dim list As ArrayList
    Dim index As Integer

    If (parameters IsNot Nothing) Then
      list = New ArrayList
      For Each m As Match In c_ParameterRegex.Matches(sql)
        If (Integer.TryParse(m.Value.Substring(6), index)) Then
          list.Add(parameters(index))
        End If
      Next
      ReorderParameters = list.ToArray
    Else
      ReorderParameters = Nothing
    End If
  End Function

  ''' <summary>
  ''' Protokolliert die Ausführdauer eines SQL-Statements.
  ''' </summary>
  ''' <param name="start">Startzeitpunkt des SQL-Statements in Ticks</param>
  ''' <param name="sql">zu protokollierendes SQL-Statement</param>
  ''' <param name="parameters">zu protokollierende Parameter</param>
  Private Shared Sub LogSQLTiming(ByVal start As Long, ByVal sql As String, ByVal ParamArray parameters() As Object)
    If (LogFacade.IsDebugEnabled) Then
      LogFacade.LogDebug(c_TraceSource, String.Format(CultureInfo.InvariantCulture, My.Resources.StorageHandler_Log_SQLTiming, sql, Join(parameters, ";"), New TimeSpan(Now.Ticks - start).ToString))
    End If
  End Sub

  ''' <summary>
  ''' Protokolliert eine SQL-Statement und dessen Parameter.
  ''' </summary>
  ''' <param name="sql">zu protokollierendes SQL-Statement</param>
  ''' <param name="parameters">zu protokollierende Parameter</param>
  Private Shared Sub LogSQL(ByVal sql As String, ByVal ParamArray parameters() As Object)
    If (LogFacade.IsInfoEnabled) Then
      LogFacade.LogInfo(c_TraceSource, String.Format(CultureInfo.InvariantCulture, My.Resources.StorageHandler_Log_SQL, sql, Join(parameters, ";")))
    End If
  End Sub

  ''' <summary>
  ''' Erstellt eine neue Dokumenten-SmeexId.
  ''' </summary>
  ''' <returns>Neue SmeexId</returns>
  Private Shared Function GetNewDocumentId() As String
    ' ACHTUNG!
    ' Die Dokumenten-SmeexId muss kundenspezifisch verwaltet werden!
    ' Es muss sichergestellt sein, dass die generierten SmeexIds eindeutig sind!
    ' Folgender Demo-Code ist dafür NICHT geeignet!
    GetNewDocumentId = String.Format(CultureInfo.InvariantCulture, "{0}.7.{1}.{2}.{3}", _
                            c_RootSmeexId, _
                            Math.Abs(My.Computer.Name.GetHashCode), _
                            Math.Abs(My.User.Name.GetHashCode), _
                            Now.Ticks)
  End Function

  ''' <summary>
  ''' Erstellt einen neuen Protokoll-Eintrag mit den übergebenen Werten.
  ''' </summary>
  ''' <param name="profile">Profil-Konfiguration</param>
  ''' <param name="action">Vorgang (EXPORT oder IMPORT)</param>
  Private Sub StartProtocoll(ByVal profile As ProfileConfiguration, ByVal action As String)
    Dim sql As String
    Dim dt As DataTable

    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    sql = "INSERT INTO SMEEX_Protokoll (Vorgang, Profil, Start) VALUES (@param0, @param1, Now())"
    ExecuteSQL(sql, action, profile.Name)
    ' @@IDENTITIY darf aufgrund von Protokoll-Einträgen nicht verwendet werden.
    ' @@SCOPE_IDENTITY wird von MS Access nicht unterstützt.
    sql = "SELECT MAX(ID) FROM SMEEX_Protokoll"
    dt = ExecuteReader(sql)
    profile.Parameter(c_ProtocolParameter) = CInt(dt.Rows(0).Item(0))
  End Sub

  ''' <summary>
  ''' Aktualisiert einen bestehenden Protokoll-Eintrag.
  ''' </summary>
  ''' <param name="profile">Profil-Konfiguration</param>
  Private Sub FinishProtocoll(ByVal profile As ProfileConfiguration, ByVal success As Boolean)
    Dim sql As String

    If (profile Is Nothing) Then Throw New ArgumentNullException("profile")
    sql = "UPDATE SMEEX_Protokoll SET Ende = Now(), Erfolgreich = @param0 WHERE (ID = @param1)"
    ExecuteSQL(sql, success, profile.Parameter(c_ProtocolParameter))
    profile.Parameter(c_ProtocolParameter) = Nothing
  End Sub

  ''' <summary>
  ''' Fügt der angegebenen Tabelle die Fehlerinformationen aus Datentabelle hinzu.
  ''' </summary>
  ''' <param name="ex">Exception, der die Informationen hinzugefügt werden sollen</param>
  ''' <param name="dt">Datentabelle, deren Fehler weitergegeben werden sollen</param>
  Private Shared Sub AppendTableError(ByVal ex As Exception, ByVal dt As DataTable)
    Dim name As String
    Dim lv As Integer

    If (Not dt Is Nothing) Then
      name = dt.TableName
      If (String.IsNullOrEmpty(name)) Then name = "tbl"
      For Each dr As DataRow In dt.Rows
        If (dr.HasErrors) Then ex.Data.Add(String.Format(CultureInfo.InvariantCulture, "{0}.{1}", name, lv), dr.RowError)
        lv += 1
      Next
    End If
  End Sub

  ''' <summary>
  ''' Konvertiert Werte zwischen Datentypen, sofern dies für ein Table-Merge notwendig ist.
  ''' </summary>
  ''' <param name="originalTable">Tabelle, die die Struktur vorgibt, resp. in die die Daten gemerged werden sollen</param>
  ''' <param name="newTable">Tabelle, dessen Daten-Typen sofern notwendig angepasst werden sollen</param>
  ''' <remarks>Es werden nur Änderungen durchgeführt, wenn in <paramref name="newTable"/> auch Daten vorhanden sind.</remarks>
  Private Shared Sub ConvertDataTypes(ByVal originalTable As DataTable, ByVal newTable As DataTable)
    Dim newCol As DataColumn

    If (originalTable IsNot Nothing) AndAlso (newTable IsNot Nothing) AndAlso (newTable.Rows.Count > 0) AndAlso (originalTable IsNot newTable) Then
      For Each col As DataColumn In originalTable.Columns
        If (newTable.Columns.Contains(col.ColumnName)) Then
          newCol = Nothing

          ' Konvertierungen falls notwendig durchführen
          If (GetType(TimeSpan).IsAssignableFrom(col.DataType)) Then
            newCol = ConvertTimeSpan(col.ColumnName, newTable)
          ElseIf (GetType(Boolean).IsAssignableFrom(col.DataType)) Then
            newCol = ConvertBoolean(col.ColumnName, newTable)
          End If

          ' Spalten falls notwendig tauschen
          If (newCol IsNot Nothing) Then
            newTable.Columns.Remove(col.ColumnName)
            newCol.ColumnName = col.ColumnName
          End If
        End If
      Next
    End If
  End Sub

  ''' <summary>
  ''' Konvertiert die Spalte <paramref name="columnName"/> in der Tabelle <paramref name="newTable"/> in eine TimeSpan-Spalte.
  ''' </summary>
  ''' <param name="columnName">Name der zu konvertierenden Spalte (muss vorhanden sein)</param>
  ''' <param name="newTable">Tabelle, dessen Spalte konvertiert werden soll</param>
  ''' <returns>Nothing oder konvertierte Spalte</returns>
  ''' <remarks>Es können nur Date-Spalten mit dieser Methode konvertiert werden.</remarks>
  Private Shared Function ConvertTimeSpan(ByVal columnName As String, ByVal newTable As DataTable) As DataColumn
    Dim oldCol As DataColumn

    ConvertTimeSpan = Nothing
    oldCol = newTable.Columns(columnName)
    If (GetType(Date).IsAssignableFrom(oldCol.DataType)) Then
      ' neue Spalte erstellen
      ConvertTimeSpan = newTable.Columns.Add(Nothing, GetType(TimeSpan))

      ' Daten kopieren
      For Each dr As DataRow In newTable.Rows
        If (Not DBNull.Value.Equals(dr(oldCol.Ordinal))) Then
          dr(ConvertTimeSpan.Ordinal) = CDate(dr(oldCol.Ordinal)).TimeOfDay
        End If
      Next
    End If
  End Function

  ''' <summary>
  ''' Konvertiert die Spalte <paramref name="columnName"/> in der Tabelle <paramref name="newTable"/> in eine Boolean-Spalte.
  ''' </summary>
  ''' <param name="columnName">Name der zu konvertierenden Spalte (muss vorhanden sein)</param>
  ''' <param name="newTable">Tabelle, dessen Spalte konvertiert werden soll</param>
  ''' <returns>Nothing oder konvertierte Spalte</returns>
  ''' <remarks>Die Werte in der Tabelle müssen DBNull oder in Boolean konvertierbar sein.</remarks>
  Private Shared Function ConvertBoolean(ByVal columnName As String, ByVal newtable As DataTable) As DataColumn
    Dim oldCol As DataColumn

    ConvertBoolean = Nothing
    oldCol = newtable.Columns(columnName)
    If (Not GetType(Boolean).IsAssignableFrom(oldCol.DataType)) Then
      ' neue Spalte erstellen
      ConvertBoolean = newtable.Columns.Add(Nothing, GetType(Boolean))

      ' Daten kopieren
      For Each dr As DataRow In newtable.Rows
        If (Not DBNull.Value.Equals(dr(oldCol.Ordinal))) Then
          dr(ConvertBoolean.Ordinal) = CBool(dr(oldCol.Ordinal))
        End If
      Next
    End If
  End Function
#End Region

#Region "Script-Methoden"
  Private Sub ExecuteSub_RegisterSmeexKey(ByVal data As ExportData, ByVal ParamArray params() As Object)
    Dim dt As DataTable
    Dim table As String
    Dim key As Integer

    If (params Is Nothing) Then Throw New ArgumentNullException("params")
    If (params.Length < 2) Then Throw New ArgumentOutOfRangeException("params")

    table = CStr(params(0))
    key = Integer.Parse(CStr(params(1)), CultureInfo.InvariantCulture)
    dt = ExecuteReader("SELECT DBKey FROM SMEEX_Key WHERE (Tabelle = @param0) AND (DBKey = @param1)", table, key)
    If (dt.Rows.Count = 0) Then
      ExecuteSQL("INSERT INTO SMEEX_Key (Tabelle, DBKey) VALUES (@param0, @param1)", table, key)
      ExecuteSub_CheckSmeexKeys(data)
    End If
  End Sub

  Private Sub ExecuteSub_CheckSmeexKeys(ByVal data As ExportData, ByVal ParamArray params() As Object)
    Dim dt As DataTable
    Dim current As Integer = Integer.MinValue
    Dim dbKey As Integer
    Dim tab As String

    dt = ExecuteReader("SELECT DBKey, Tabelle FROM SMEEX_Key WHERE (SmeexKey IS NULL)")
    For Each dr As DataRow In dt.Rows
      ' aktuelles Maximum ermitteln
      dbKey = CInt(dr("DBKey"))
      tab = CStr(dr("Tabelle"))
      If (current <> Integer.MinValue) Then
        current += 1
      Else
        dt = ExecuteReader("SELECT MAX(SmeexKey) FROM SMEEX_Key")
        If (dt.Rows.Count < 1) OrElse (DBNull.Value.Equals(dt.Rows(0)(0))) Then
          current = 1
        Else
          current = CInt(dt.Rows(0)(0)) + 1
        End If
      End If

      ' neuer PK setzen
      ExecuteSQL("UPDATE SMEEX_Key SET SmeexKey = @param0 WHERE (Tabelle = @param1) AND (DBKey = @param2)", current, tab, dbKey)
    Next
  End Sub
#End Region

#Region " IDisposable Support "
  Private disposedValue As Boolean     ' So ermitteln Sie überflüssige Aufrufe

  ' IDisposable
  Protected Overrides Sub Dispose(ByVal disposing As Boolean)
    If Not Me.disposedValue Then
      If disposing Then
        ' Anderen Zustand freigeben (verwaltete Objekte).
        If (m_Connection IsNot Nothing) Then
          m_Connection.Close()
          m_Connection.Dispose()
          m_Connection = Nothing
        End If
        LogFacade.Unregister(Me)
      End If

      ' Eigenen Zustand freigeben (nicht verwaltete Objekte).
      ' Große Felder auf NULL festlegen.
    End If
    Me.disposedValue = True
    MyBase.Dispose(disposing)
  End Sub
#End Region

#Region "TraceListener"
  ''' <summary>
  ''' Protokolliert die Log-Ausgaben in der Datenbank.
  ''' </summary>
  Private Sub WriteLog(ByVal dateTime As Date, ByVal category As String, ByVal source As String, ByVal message As String, ByVal stackTrace As String)
    Dim sql As String

    sql = "INSERT INTO SMEEX_Log ([DatumZeit], [Kategorie], [Quelle], [Meldung], [StackTrace]) VALUES (@param0, @param1, @param2, @param3, @param4)"
    Try
      ' das übergebene DateTime-Objekt entspricht einer UTC-Zeitangabe!
      InternalExecuteSQL(sql, dateTime.ToLocalTime.ToString("G", CultureInfo.InvariantCulture), category, source, message, stackTrace)
    Catch ex As Exception
      ' Alle Fehler ignorieren!
    End Try
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.Write" />
  ''' </summary>
  Public Overloads Overrides Sub Write(ByVal message As String)
    Write(message, String.Empty)
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.Write" />
  ''' </summary>
  Public Overrides Sub Write(ByVal message As String, ByVal category As String)
    WriteLog(Date.Now, category, String.Empty, message, String.Empty)
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.WriteLine" />
  ''' </summary>
  Public Overloads Overrides Sub WriteLine(ByVal message As String)
    Write(message)
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.WriteLine" />
  ''' </summary>
  Public Overrides Sub WriteLine(ByVal message As String, ByVal category As String)
    Write(message, category)
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.TraceData" />
  ''' </summary>
  Public Overrides Sub TraceData(ByVal eventCache As TraceEventCache, ByVal source As String, ByVal eventType As TraceEventType, ByVal id As Integer, ByVal ParamArray data As Object())
    If (eventCache Is Nothing) Then Throw New ArgumentNullException("eventCache")
    If ((Me.Filter Is Nothing) OrElse Me.Filter.ShouldTrace(eventCache, source, eventType, id, Nothing, Nothing, Nothing, data)) Then
      Dim builder As New StringBuilder
      If (Not data Is Nothing) Then
        Dim i As Integer
        For i = 0 To data.Length - 1
          If (i <> 0) Then builder.Append(", ")
          If (Not data(i) Is Nothing) Then builder.Append(data(i).ToString)
        Next i
      End If
      WriteLog(eventCache.DateTime, eventType.ToString, source, builder.ToString, eventCache.Callstack)
    End If
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.TraceData" />
  ''' </summary>
  Public Overrides Sub TraceData(ByVal eventCache As TraceEventCache, ByVal source As String, ByVal eventType As TraceEventType, ByVal id As Integer, ByVal data As Object)
    If (eventCache Is Nothing) Then Throw New ArgumentNullException("eventCache")
    If ((Me.Filter Is Nothing) OrElse Me.Filter.ShouldTrace(eventCache, source, eventType, id, Nothing, Nothing, data, Nothing)) Then
      Dim message As String = String.Empty
      If (Not data Is Nothing) Then message = data.ToString
      WriteLog(eventCache.DateTime, eventType.ToString, source, message, eventCache.Callstack)
    End If
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.TraceEvent" />
  ''' </summary>
  Public Overrides Sub TraceEvent(ByVal eventCache As TraceEventCache, ByVal source As String, ByVal eventType As TraceEventType, ByVal id As Integer, ByVal message As String)
    If (eventCache Is Nothing) Then Throw New ArgumentNullException("eventCache")
    If ((Me.Filter Is Nothing) OrElse Me.Filter.ShouldTrace(eventCache, source, eventType, id, message, Nothing, Nothing, Nothing)) Then
      WriteLog(eventCache.DateTime, eventType.ToString, source, message, eventCache.Callstack)
    End If
  End Sub

  ''' <summary>
  ''' <see cref="TraceListener.TraceEvent" />
  ''' </summary>
  Public Overrides Sub TraceEvent(ByVal eventCache As TraceEventCache, ByVal source As String, ByVal eventType As TraceEventType, ByVal id As Integer, ByVal format As String, ByVal ParamArray args As Object())
    If (eventCache Is Nothing) Then Throw New ArgumentNullException("eventCache")
    If ((Me.Filter Is Nothing) OrElse Me.Filter.ShouldTrace(eventCache, source, eventType, id, format, args, Nothing, Nothing)) Then
      If (Not args Is Nothing) Then format = String.Format(CultureInfo.InvariantCulture, format, args)
      WriteLog(eventCache.DateTime, eventType.ToString, source, format, eventCache.Callstack)
    End If
  End Sub
#End Region

#Region "Hilfsklassen"
  Private Class ExportData
    Implements IDisposable
#Region "Variablen"
    Private m_Profile As ProfileConfiguration
    Private m_Archive As Archive
    Private m_TableDefinition As Table
    Private m_CurrentData As DataTable
    Private m_RelationData As DataTable
    Private m_SourceRelation As Relation
#End Region

#Region "Konstruktoren"
    Public Sub New(ByVal profile As ProfileConfiguration, ByVal arch As Archive, ByVal tableDefinition As Table, ByVal currentData As DataTable, ByVal relationData As DataTable, ByVal sourceRelation As Relation)
      m_Profile = profile
      m_Archive = arch
      m_TableDefinition = tableDefinition
      m_CurrentData = currentData
      m_RelationData = relationData
      m_SourceRelation = sourceRelation
    End Sub
#End Region

#Region "Properties"
    Public ReadOnly Property Profile() As ProfileConfiguration
      Get
        Profile = m_Profile
      End Get
    End Property
    Public ReadOnly Property Archive() As Archive
      Get
        Archive = m_Archive
      End Get
    End Property
    Public ReadOnly Property TableDefinition() As Table
      Get
        TableDefinition = m_TableDefinition
      End Get
    End Property
    Public ReadOnly Property CurrentData() As DataTable
      Get
        CurrentData = m_CurrentData
      End Get
    End Property
    Public ReadOnly Property RelationData() As DataTable
      Get
        RelationData = m_RelationData
      End Get
    End Property
    Public ReadOnly Property SourceRelation() As Relation
      Get
        SourceRelation = m_SourceRelation
      End Get
    End Property
#End Region

#Region " IDisposable Support "
    Private disposedValue As Boolean

    ' IDisposable
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
      If Not Me.disposedValue Then
        If disposing Then
          ' Anderen Zustand freigeben (verwaltete Objekte).
          m_Profile = Nothing
          m_Archive = Nothing
          m_TableDefinition = Nothing
          m_CurrentData = Nothing
          m_RelationData = Nothing
          m_SourceRelation = Nothing
        End If

        ' Eigenen Zustand freigeben (nicht verwaltete Objekte).
        ' Große Felder auf NULL festlegen.
      End If
      Me.disposedValue = True
    End Sub

    ' Dieser Code wird von Visual Basic hinzugefügt, um das Dispose-Muster richtig zu implementieren.
    Public Sub Dispose() Implements IDisposable.Dispose
      ' Ändern Sie diesen Code nicht. Fügen Sie oben in Dispose(ByVal disposing As Boolean) Bereinigungscode ein.
      Dispose(True)
      GC.SuppressFinalize(Me)
    End Sub
#End Region
  End Class
#End Region
End Class
